@@ -13,7 +13,7 @@
  *  * Neither the name of the ALICE Project-Team nor the names of its
  *  contributors may be used to endorse or promote products derived from this
  *  software without specific prior written permission.
- * 
+ *
  *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
@@ -48,7 +48,7 @@ namespace GEO {
     void AttributeStoreObserver::register_me(AttributeStore* store) {
         store->register_observer(this);
     }
-    
+
     void AttributeStoreObserver::unregister_me(AttributeStore* store) {
         store->unregister_observer(this);
     }
@@ -58,33 +58,31 @@ namespace GEO {
     AttributeStoreCreator::~AttributeStoreCreator() {
     }
 
-    /******************************************************************/    
-    
+    /******************************************************************/
+
     std::map<std::string, AttributeStoreCreator_var>
-         AttributeStore::type_name_to_creator_;
+    AttributeStore::type_name_to_creator_;
 
     std::map<std::string, std::string>
-         AttributeStore::typeid_name_to_type_name_;
+    AttributeStore::typeid_name_to_type_name_;
 
     std::map<std::string, std::string>
-         AttributeStore::type_name_to_typeid_name_;
-    
-    AttributeStore::AttributeStore(
-        index_t elemsize,
-        index_t dim
-    ) :
+    AttributeStore::type_name_to_typeid_name_;
+
+    AttributeStore::AttributeStore(size_t elemsize, index_t dim) :
         element_size_(elemsize),
         dimension_(dim),
         cached_base_addr_(nullptr),
         cached_size_(0),
-	cached_capacity_(0),
+        cached_capacity_(0),
         lock_(GEOGRAM_SPINLOCK_INIT)
     {
     }
-    
+
     void AttributeStore::notify(
         Memory::pointer base_addr, index_t size, index_t dim
     ) {
+        Process::acquire_spinlock(lock_);
         if(
             size != cached_size_ ||
             base_addr != cached_base_addr_ ||
@@ -93,20 +91,21 @@ namespace GEO {
             cached_base_addr_ = base_addr;
             cached_size_ = size;
             dimension_ = dim;
-	    for(auto cur : observers_) {
-		cur->notify(cached_base_addr_, cached_size_, dim);
-	    }
+            for(auto cur : observers_) {
+                cur->notify(cached_base_addr_, cached_size_, dim);
+            }
         }
+        Process::release_spinlock(lock_);
     }
-    
+
     AttributeStore::~AttributeStore() {
-	// Disconnect all the attributes, for the special case where
-	// the AttributeStore is destroyed before the Attributes, can
-	// occur for instance when using Lua scripting with Attribute wrapper
-	// objects.
-	for(auto cur : observers_) {
-	    cur->disconnect();
-	}
+        // Disconnect all the attributes, for the special case where
+        // the AttributeStore is destroyed before the Attributes, can
+        // occur for instance when using Lua scripting with Attribute wrapper
+        // objects.
+        for(auto cur : observers_) {
+            cur->disconnect();
+        }
     }
 
     void AttributeStore::register_observer(AttributeStoreObserver* observer) {
@@ -114,71 +113,132 @@ namespace GEO {
         geo_assert(observers_.find(observer) == observers_.end());
         observers_.insert(observer);
         observer->notify(cached_base_addr_, cached_size_, dimension_);
-        Process::release_spinlock(lock_);        
+        Process::release_spinlock(lock_);
     }
 
     void AttributeStore::unregister_observer(AttributeStoreObserver* observer) {
         Process::acquire_spinlock(lock_);
-	auto it = observers_.find(observer);
+        auto it = observers_.find(observer);
         geo_assert(it != observers_.end());
         observers_.erase(it);
-        Process::release_spinlock(lock_);                
+        Process::release_spinlock(lock_);
     }
 
-    void AttributeStore::apply_permutation(
-        const vector<index_t>& permutation
+    void AttributeStore::apply_permutation_with_lifecycle(
+	const vector<index_t>& permutation_in
     ) {
+	Memory::pointer pdata = cached_base_addr_;
+	vector<index_t>& perm = const_cast<vector<index_t>&>(permutation_in);
+	geo_debug_assert(Permutation::is_valid(perm));
+	size_t item_size = element_size_ * dimension_;
+	Memory::pointer temp = static_cast<Memory::pointer>(alloca(item_size));
+	for(index_t k = 0; k < perm.size(); ++k) {
+	    if(Permutation::is_marked(perm, k)) {
+		continue;
+	    }
+	    index_t i = k;
+	    index_t j = perm[k];
+
+	    lifecycle_->copy_construct_array(
+		temp, pdata + i * item_size, dimension_
+	    );
+
+	    Permutation::mark(perm, k);
+	    while(j != k) {
+		lifecycle_->assign_array(
+		    pdata + i * item_size, pdata + j * item_size, dimension_
+		);
+		index_t nj = perm[j];
+		Permutation::mark(perm, j);
+		i = j;
+		j = nj;
+	    }
+	    lifecycle_->assign_array(pdata + i * item_size, temp, dimension_);
+	    lifecycle_->destroy_array(temp, dimension_);
+	}
+	for(index_t k = 0; k < perm.size(); ++k) {
+	    Permutation::unmark(perm, k);
+	}
+    }
+
+    void AttributeStore::apply_permutation(const vector<index_t>& permutation) {
         geo_debug_assert(permutation.size() <= cached_size_);
-        Permutation::apply(
-            cached_base_addr_, permutation, element_size_ * dimension_
-        );
+	if(lifecycle_.is_null()) {
+	    Permutation::apply(
+		cached_base_addr_, permutation, element_size_ * dimension_
+	    );
+	} else {
+	    apply_permutation_with_lifecycle(permutation);
+	}
     }
 
-    void AttributeStore::compress(
-        const vector<index_t>& old2new
-    ) {
+    void AttributeStore::compress(const vector<index_t>& old2new) {
         geo_debug_assert(old2new.size() <= cached_size_);
-        index_t item_size = element_size_ * dimension_;
-        for(index_t i=0; i<old2new.size(); ++i) {
-            index_t j = old2new[i];
-            if(j == index_t(-1) || j == i) {
-                continue;
-            }
-            geo_debug_assert(j <= i);
-            Memory::copy(
-                cached_base_addr_+j*item_size,
-                cached_base_addr_+i*item_size,
-                item_size
-            );
-        }
+        size_t item_size = size_t(element_size_) * dimension_;
+	for(index_t i=0; i<old2new.size(); ++i) {
+	    index_t j = old2new[i];
+	    if(j == NO_INDEX || j == i) {
+		continue;
+	    }
+	    geo_debug_assert(j <= i);
+	    if(lifecycle_.is_null())  {
+		Memory::copy(
+		    cached_base_addr_+size_t(j)*item_size,
+		    cached_base_addr_+size_t(i)*item_size,
+		    item_size
+		);
+	    } else {
+		lifecycle_->assign_array(
+		    cached_base_addr_+size_t(j)*item_size,
+		    cached_base_addr_+size_t(i)*item_size,
+		    dimension_
+		);
+	    }
+	}
     }
 
     void AttributeStore::zero() {
-        Memory::clear(
-            cached_base_addr_, element_size_ * dimension_ * cached_size_
-        );
+	if(lifecycle_.is_null()) {
+	    Memory::clear(
+		cached_base_addr_, element_size_ * dimension_ * cached_size_
+	    );
+	} else {
+	    index_t nb_elements = cached_size_ * dimension_;
+	    for(index_t i=0; i<nb_elements; ++i) {
+		lifecycle_->reset(cached_base_addr_ + i * element_size_);
+	    }
+	}
     }
 
     void AttributeStore::swap_items(index_t i, index_t j) {
         geo_debug_assert(i < cached_size_);
         geo_debug_assert(j < cached_size_);
-        index_t item_size = element_size_ * dimension_;
-        void* temp = alloca(item_size);
-        Memory::copy(
-            temp,
-            cached_base_addr_+i*item_size,
-            item_size
-        );
-        Memory::copy(
-            cached_base_addr_+i*item_size,
-            cached_base_addr_+j*item_size,
-            item_size
-        );
-        Memory::copy(
-            cached_base_addr_+j*item_size,
-            temp,
-            item_size
-        );
+	size_t item_size = element_size_ * dimension_;
+	if(lifecycle_.is_null()) {
+	    void* temp = alloca(item_size);
+	    Memory::copy(
+		temp,
+		cached_base_addr_+i*item_size,
+		item_size
+	    );
+	    Memory::copy(
+		cached_base_addr_+i*item_size,
+		cached_base_addr_+j*item_size,
+		item_size
+	    );
+	    Memory::copy(
+		cached_base_addr_+j*item_size,
+		temp,
+		item_size
+	    );
+	} else {
+	    for(index_t c=0; c<dimension_; ++c) {
+		lifecycle_->swap(
+		    cached_base_addr_+i*item_size+c*element_size(),
+		    cached_base_addr_+j*item_size+c*element_size()
+		);
+	    }
+	}
     }
 
     void AttributeStore::scale_item(index_t to, double s) {
@@ -191,7 +251,29 @@ namespace GEO {
         geo_argused(s);
         geo_argused(from);
     }
-    
+
+    void AttributeStore::register_attribute_creator(
+        AttributeStoreCreator* creator,
+        const std::string& element_type_name,
+        const std::string& element_typeid_name
+    ) {
+        if(element_type_name_is_known(element_type_name)) {
+            Logger::warn("Attributes") << element_type_name
+                                       << " already registered"
+                                       << std::endl;
+            if(element_typeid_name_is_known(element_typeid_name)) {
+                bool already_registered_attribute_has_same_type = (
+                    type_name_to_typeid_name_[element_type_name] ==
+                    element_typeid_name
+                );
+                geo_assert(already_registered_attribute_has_same_type);
+            }
+        }
+        type_name_to_creator_[element_type_name] = creator;
+        typeid_name_to_type_name_[element_typeid_name] = element_type_name;
+        type_name_to_typeid_name_[element_type_name] = element_typeid_name;
+    }
+
     /*************************************************************************/
 
     AttributesManager::AttributesManager() : size_(0), capacity_(0) {
@@ -200,14 +282,14 @@ namespace GEO {
     AttributesManager::~AttributesManager() {
         clear(false,false);
     }
-    
+
     void AttributesManager::resize(index_t new_size) {
         if(new_size == size_) {
             return;
         }
-	for(auto& cur : attributes_) {
-	    cur.second->resize(new_size);
-	}
+        for(auto& cur : attributes_) {
+            cur.second->resize(new_size);
+        }
         size_ = new_size;
     }
 
@@ -215,35 +297,35 @@ namespace GEO {
         if(new_capacity <= capacity_) {
             return;
         }
-	for(auto& cur : attributes_) {
-	    cur.second->reserve(new_capacity);
-	}
+        for(auto& cur : attributes_) {
+            cur.second->reserve(new_capacity);
+        }
         capacity_ = new_capacity;
     }
-    
+
     void AttributesManager::apply_permutation(
         const vector<index_t>& permutation
     ) {
-	for(auto& cur : attributes_) {
-	    cur.second->apply_permutation(permutation);	    
-	}
+        for(auto& cur : attributes_) {
+            cur.second->apply_permutation(permutation);
+        }
     }
 
     void AttributesManager::compress(
         const vector<index_t>& old2new
     ) {
-	for(auto& cur : attributes_) {
-	    cur.second->compress(old2new);	    
-	}
+        for(auto& cur : attributes_) {
+            cur.second->compress(old2new);
+        }
     }
-    
-    
+
+
     void AttributesManager::bind_attribute_store(
         const std::string& name, AttributeStore* as
     ) {
         geo_assert(find_attribute_store(name) == nullptr);
         attributes_[name] = as;
-	as->reserve(capacity_);
+        as->reserve(capacity_);
         as->resize(size_);
     }
 
@@ -251,11 +333,11 @@ namespace GEO {
         vector<std::string>& names
     ) const {
         names.clear();
-	for(auto& cur : attributes_) {
-	    names.push_back(cur.first);
-	}
+        for(auto& cur : attributes_) {
+            names.push_back(cur.first);
+        }
     }
-    
+
     AttributeStore* AttributesManager::find_attribute_store(
         const std::string& name
     ) {
@@ -269,16 +351,16 @@ namespace GEO {
     const AttributeStore* AttributesManager::find_attribute_store(
         const std::string& name
     ) const {
-	auto it = attributes_.find(name);
+        auto it = attributes_.find(name);
         if(it == attributes_.end()) {
             return nullptr;
         }
         return it->second;
     }
-    
+
 
     void AttributesManager::delete_attribute_store(const std::string& name) {
-	auto it = attributes_.find(name);
+        auto it = attributes_.find(name);
         geo_assert(it != attributes_.end());
         geo_assert(!it->second->has_observers());
         delete it->second;
@@ -299,63 +381,63 @@ namespace GEO {
 
     void AttributesManager::clear(bool keep_attributes, bool keep_memory) {
         if(keep_attributes) {
-	    for(auto& cur : attributes_) {
+            for(auto& cur : attributes_) {
                 cur.second->clear(keep_memory);
             }
         } else {
-	    for(auto& cur : attributes_) {
-		delete cur.second;
-	    }
+            for(auto& cur : attributes_) {
+                delete cur.second;
+            }
             attributes_.clear();
         }
         size_ = 0;
     }
 
     void AttributesManager::zero() {
-	for(auto& cur : attributes_) {
-	    cur.second->zero();
-	}
+        for(auto& cur : attributes_) {
+            cur.second->zero();
+        }
     }
 
     void AttributesManager::copy(const AttributesManager& rhs) {
         clear(false, false);
-	reserve(rhs.capacity());
+        reserve(rhs.capacity());
         resize(rhs.size());
-	for(auto& cur : rhs.attributes_) {
-            bind_attribute_store(cur.first, cur.second->clone());	    
-	}
+        for(auto& cur : rhs.attributes_) {
+            bind_attribute_store(cur.first, cur.second->clone());
+        }
     }
 
     void AttributesManager::copy_item(index_t to, index_t from) {
-	for(auto& cur : attributes_) {
-	    cur.second->copy_item(to,from);
-	}
+        for(auto& cur : attributes_) {
+            cur.second->copy_item(to,from);
+        }
     }
 
     void AttributesManager::swap_items(index_t i, index_t j) {
-	for(auto& cur : attributes_) {
-	    cur.second->swap_items(i,j);
-	}
+        for(auto& cur : attributes_) {
+            cur.second->swap_items(i,j);
+        }
     }
 
     void AttributesManager::zero_item(index_t i) {
-	for(auto& cur : attributes_) {
-	    cur.second->zero_item(i);
-	}
+        for(auto& cur : attributes_) {
+            cur.second->zero_item(i);
+        }
     }
-    
+
     void AttributesManager::scale_item(index_t i, double s) {
-	for(auto& cur : attributes_) {
-	    cur.second->scale_item(i,s);
-	}
+        for(auto& cur : attributes_) {
+            cur.second->scale_item(i,s);
+        }
     }
 
     void AttributesManager::madd_item(index_t i, double s, index_t j) {
-	for(auto& cur : attributes_) {
-	    cur.second->madd_item(i,s,j);
-	}
+        for(auto& cur : attributes_) {
+            cur.second->madd_item(i,s,j);
+        }
     }
-    
+
     bool AttributesManager::copy_attribute(
         const std::string& name, const std::string& new_name
     ) {
@@ -380,10 +462,20 @@ namespace GEO {
             ) {
                 return false;
             }
-            memcpy(
-                new_store->data(), store->data(),
-                store->size() * store->dimension() * store->element_size()
-            );
+
+	    geo_debug_assert(store->lifecycle() == new_store->lifecycle());
+	    if(store->lifecycle() == nullptr) {
+		memcpy(
+		    new_store->data(), store->data(),
+		    store->size() * store->dimension() * store->element_size()
+		);
+	    } else {
+		store->lifecycle()->assign_array(
+		    Memory::pointer(new_store->data()),
+		    Memory::pointer(store->data()),
+		    store->size() * store->dimension()
+		);
+	    }
         } else {
             AttributeStore* new_store = store->clone();
             attributes_[new_name] = new_store;
@@ -408,8 +500,8 @@ namespace GEO {
         return true;
     }
 
-    
-    /************************************************************************/ 
+
+    /************************************************************************/
 
     index_t ScalarAttributeAdapterBase::nb_scalar_elements_per_item(
         const AttributeStore* store
@@ -445,14 +537,14 @@ namespace GEO {
         if(pos != std::string::npos) {
             try {
                 if(pos+2 > name.length()) {
-                    result = index_t(-1);
+                    result = NO_INDEX;
                 } else {
                     result = String::to_uint(
                         name.substr(pos+1, name.length()-pos-2)
                     );
                 }
             } catch(...) {
-                result = index_t(-1);
+                result = NO_INDEX;
             }
         }
         return result;
@@ -481,21 +573,21 @@ namespace GEO {
 
         if(
             store->element_typeid_name() == typeid(Numeric::int32).name() ||
-            store->element_typeid_name() == typeid(int).name() 
+            store->element_typeid_name() == typeid(int).name()
         ) {
-            return ET_INT32;                
+            return ET_INT32;
         }
 
         if(
             store->element_typeid_name() == typeid(Numeric::float32).name() ||
-            store->element_typeid_name() == typeid(float).name() 
+            store->element_typeid_name() == typeid(float).name()
         ) {
             return ET_FLOAT32;
         }
 
         if(
             store->element_typeid_name() == typeid(Numeric::float64).name() ||
-            store->element_typeid_name() == typeid(double).name() 
+            store->element_typeid_name() == typeid(double).name()
         ) {
             return ET_FLOAT64;
         }
@@ -507,10 +599,10 @@ namespace GEO {
         if(store->element_typeid_name() == typeid(vec3).name()) {
             return ET_VEC3;
         }
-        
+
         return ET_NONE;
     }
-    
+
     void ScalarAttributeAdapterBase::bind_if_is_defined(
         const AttributesManager& manager, const std::string& name
     ) {
@@ -519,9 +611,9 @@ namespace GEO {
         element_index_ = attribute_element_index(name);
         store_ = manager_->find_attribute_store(attribute_base_name(name));
 
-        if(store_ == nullptr || element_index_ == index_t(-1)) {
+        if(store_ == nullptr || element_index_ == NO_INDEX) {
             store_ = nullptr;
-            element_index_ = index_t(-1);
+            element_index_ = NO_INDEX;
             return;
         }
 
@@ -529,7 +621,7 @@ namespace GEO {
 
         if(element_type_ == ET_NONE) {
             store_ = nullptr;
-            element_index_ = index_t(-1);
+            element_index_ = NO_INDEX;
             return;
         }
 
@@ -538,12 +630,12 @@ namespace GEO {
         // or 3*store's dimension if a vec3)
         if(element_index_ >= nb_scalar_elements_per_item(store_)) {
             store_ = nullptr;
-            element_index_ = index_t(-1);
+            element_index_ = NO_INDEX;
             element_type_ = ET_NONE;
-	    return;
+            return;
         }
-        
-        register_me(const_cast<AttributeStore*>(store_));                
+
+        register_me(const_cast<AttributeStore*>(store_));
     }
 
     bool ScalarAttributeAdapterBase::is_defined(
@@ -553,13 +645,13 @@ namespace GEO {
         const AttributeStore* store = manager.find_attribute_store(
             attribute_name
         );
-        
+
         if(store == nullptr) {
             return false;
         }
-        
+
         index_t element_index = attribute_element_index(name);
-        if(element_index == index_t(-1)) {
+        if(element_index == NO_INDEX) {
             return false;
         }
 
@@ -570,11 +662,10 @@ namespace GEO {
         if(element_type(store) == ET_NONE) {
             return false;
         }
-        
+
         return true;
     }
-    
-    /************************************************************************/ 
-    
-}
 
+    /************************************************************************/
+
+}
